#
# Copyright 2008-2024 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Results of the Observing Blocks"""


from astropy.io import fits

from numina.datamodel import DataModel
from numina.types.dataframe import DataFrame


class ObservingBlockBase:
    def __init__(self, instrument='UNKNOWN', mode='UNKNOWN'):
        self.id = 1
        self.instrument = instrument
        self.mode = mode
        self.frames = []
        # other OBs related to this one
        self.children = []
        self.parent = None

    def get_sample_frame(self):
        """Return first available frame in observation result"""
        for frame in self.frames:
            return frame


class ObservingBlock(ObservingBlockBase):
    """
    Description of an observing block
    """

    def __init__(self, instrument='UNKNOWN', mode='UNKNOWN'):
        super().__init__(instrument, mode)
        # The results of processing children OBs
        # These values are added by method ObservingMode.build_ob
        # with a query to the stored results
        self.results = {}

        # Provide requirements for reduction
        self.requirements = {}
        # Pipeline used to process this
        self.pipeline = 'default'
        # Name and object of the instrument configuration
        self.profile = '00000000-0000-0000-0000-000000000000'
        self.configuration = 'default'
        #
        self.prodid = None
        # tags are added by method by Recipe.build_recipe_input
        self.tags = {}
        #
        self.labels = {}

    def get_sample_frame(self):
        """Return first available frame in observation result"""
        for frame in self.frames:
            return frame

        for res in self.results.values():
            return res

        return None


class ObservationResult(ObservingBlock):
    """The result of an observing block"""

    def __init__(self, instrument='UNKNOWN', mode='UNKNOWN'):
        super().__init__(instrument, mode)

    def update_with_product(self, prod):
        self.tags = prod.tags
        self.frames = [prod.content]
        self.prodid = prod.id

    @property
    def images(self):
        return self.frames

    @images.setter
    def images(self, value):
        self.frames = value

    def __str__(self):
        return 'ObservationResult(id={}, instrument={}, mode={})'.format(
            self.id,
            self.instrument,
            self.mode
        )

    def metadata_with(self, datamodel: DataModel) -> dict:
        """Extract metadata from the OB using a DataModel object"""
        origin = {}
        imginfo = datamodel.gather_info_oresult(self)
        origin['info'] = imginfo
        if imginfo:
            first = imginfo[0]
            origin["block_uuid"] = first['block_uuid']
            origin['insconf_uuid'] = first['insconf_uuid']
            # The same field
            origin['date_obs'] = first['observation_date']
            origin['observation_date'] = first['observation_date']
            # Ids of the images
            origin['frames'] = [img['imgid'] for img in imginfo]
        return origin


def dataframe_from_list(values):
    """Build a DataFrame object from a list."""
    if isinstance(values, str):
        return DataFrame(filename=values)
    elif isinstance(values, fits.HDUList):
        return DataFrame(frame=values)
    else:
        return None


def oblock_from_dict(values: dict) -> ObservingBlock:
    """Build a ObservingBlock object from a dictionary."""

    obsres = ObservingBlock()

    ikey = 'frames'
    # Workaround
    if 'images' in values:
        ikey = 'images'

    obsres.id = values.get('id', 1)
    obsres.mode = values['mode']
    obsres.instrument = values['instrument']
    # obsres.configuration = values.get('configuration', 'default')
    obsres.pipeline = values.get('pipeline', 'default')
    obsres.children = values.get('children',  [])
    obsres.parent = values.get('parent', None)
    obsres.results = values.get('results', {})
    obsres.labels = values.get('labels', {})
    obsres.requirements = values.get('requirements', {})
    try:
        obsres.frames = [dataframe_from_list(val) for val in values[ikey]]
    except Exception:
        obsres.frames = []

    return obsres


def obsres_from_dict(values: dict) -> ObservationResult:
    """Build a ObservationResult object from a dictionary."""

    obsres = ObservationResult()

    ikey = 'frames'
    # Workaround
    if 'images' in values:
        ikey = 'images'

    obsres.id = values.get('id', 1)
    obsres.mode = values['mode']
    obsres.instrument = values['instrument']
    # obsres.configuration = values.get('configuration', 'default')
    obsres.pipeline = values.get('pipeline', 'default')
    obsres.children = values.get('children',  [])
    obsres.parent = values.get('parent', None)
    obsres.results = values.get('results', {})
    obsres.labels = values.get('labels', {})
    obsres.requirements = values.get('requirements', {})
    try:
        obsres.frames = [dataframe_from_list(val) for val in values[ikey]]
    except Exception:
        obsres.frames = []

    return obsres



#
# Copyright 2018-2019 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Classes to express tags constraints in recipe requirements

Each type has a series of tag attributes inside the .tags member. Operating
with tags creates tag expressions.

>>> tags1 = Tagged(['vph', 'insmode'])
>>> tags1.tags.insmode == "LCB"

When processed, this represents that the `insmode` key of a particular calibration must be equal to "LCB".

Most of the time, you don't need a fixed value, but a value that comes from the observed data

>>> tags1.tags.insmode == tags1.tags.p_("insmode")

In this case, the `insmode` tag of the calibration must be equal to the value of insmode extracted from
the observed data.

Expresions can be nested and make use of the typical logical conectors (AND, OR, NOT), like:

>>> (tags1.tags.insmode == tags1.tags.p_("insmode")) & (tags1.tags.vph == tags1.tags.p_("vph"))

Where & represents AND, | represents OR and ~ represents NOT

"""

import operator

try:
    from functools import reduce
except ImportError:
    pass


def map_tree(visitor, tree):
    """Apply function to nodes"""
    newn = [map_tree(visitor, node) for node in tree.nodes]
    return visitor(tree, newn)


def filter_tree(condition, tree):
    """Return parts of the tree that fulfill condition"""
    if condition(tree):
        for node in tree.nodes:
            # this works in python > 3.3
            # yield from filter_tree(condition, node)
            for n in filter_tree(condition, node):
                yield n
    yield tree


class Expression(object):
    """Base class for expressions"""

    def __init__(self, *args):

        self.nodes = []
        self.metadata = {}
        for arg in args:
            if isinstance(arg, Expression):
                self.nodes.append(arg)
            else:
                self.nodes.append(ConstExpr(arg))

        self._fields = set()
        self._places = set()

        for node in self.nodes:
            self._fields.update(node.fields())
            self._places.update(node.places())

    def fields(self):
        return self._fields

    def tags(self):
        return self._places

    def places(self):
        return self._places

    def is_terminal(self):
        """True for leaf nodes"""
        return len(self.nodes) == 0

    def copy(self):

        def copy_node(node, children):
            return node.clone(children)

        return map_tree(copy_node, self)

    def __eq__(self, other):
        return PredEq(self, other)

    def __gt__(self, other):
        return PredGt(self, other)

    def __ge__(self, other):
        return PredGe(self, other)

    def __le__(self, other):
        return PredLe(self, other)

    def __lt__(self, other):
        return PredLt(self, other)

    def __ne__(self, other):
        return PredNe(self, other)

    def __or__(self, other):
        if isinstance(other, CompoundExpr):
            return PredOr(self, other)

        if isinstance(other, bool):
            if other:
                return True
            else:
                return other

        return NotImplemented

    def __ror__(self, other):
        return self.__or__(other)

    def __and__(self, other):
        if isinstance(other, CompoundExpr):
            return PredAnd(self, other)

        if isinstance(other, bool):
            if other:
                return self
            else:
                return False

        return NotImplemented

    def __rand__(self, other):
        return self.__and__(other)

    def __invert__(self):
        return PredNot(self)

    @staticmethod
    def map_tree(visitor, tree):
        return map_tree(visitor, tree)

    def fill_placeholders(self, **kwargs):
        """Substitute Placeholder nodes by its value in tags"""
        def change_p_node_tags(node, children):
            if isinstance(node, Placeholder) and (node.name in kwargs):
                return ConstExpr(kwargs[node.name])
            else:
                return node.clone(children)

        return map_tree(change_p_node_tags, self)

    fill_tags = fill_placeholders

    def eval(self, **kwargs):

        def eval_nodes(node, children):
            if isinstance(node, BinaryExpr):
                return node.operator(children[0], children[1])
            if isinstance(node, UnaryExpr):
                return node.operator(children[0])
            else:
                return node.eval(**kwargs)

        return map_tree(eval_nodes, self)

    def clone(self, children):
        raise NotImplementedError


class AtomicExpr(Expression):
    """"Atomic expression"""

    def __init__(self, name, value):
        super(AtomicExpr, self).__init__()
        self.name = name
        self.value = value


class TagRepr(AtomicExpr):
    """A representation of a Tag"""

    def __init__(self, name, metadata=None):
        super(TagRepr, self).__init__(name, name)
        self.metadata = metadata or {}
        self._fields.add(name)

    def clone(self, nodes):
        return self

    def eval(self, **kwargs):
        if self.value in kwargs:
            return kwargs[self.value]
        else:
            return self

    def __repr__(self):
        return f"TagRepr({self.name})"


class Placeholder(AtomicExpr):
    """A representation of a value expected to be substituted"""

    def __init__(self, name):
        super(Placeholder, self).__init__(name, name)
        self._places.add(name)

    def __repr__(self):
        return f"Placeholder({self.name})"

    def clone(self, nodes):
        return self


class ConstExpr(AtomicExpr):
    """A representation of a constant value"""

    def __init__(self, value):
        super(ConstExpr, self).__init__("ConstExpr", value)

    def clone(self, nodes):
        return self

    def __repr__(self):
        return f"ConstExpr({self.value})"

    def eval(self, **kwargs):
        return self.value


ConstExprTrue = ConstExpr(True)
ConstExprFalse = ConstExpr(False)


class CompoundExpr(Expression):
    """Compound expression"""
    pass


class UnaryExpr(CompoundExpr):
    def __init__(self, pred, oper):
        super(UnaryExpr, self).__init__(pred)
        self.pred = self.nodes[0]
        self.operator = oper

    def clone(self, nodes):
        new = self.__class__(nodes[0], self.operator)
        return new


class BinaryExpr(CompoundExpr):
    def __init__(self, lhs, rhs, oper, op_rep):
        super(BinaryExpr, self).__init__(lhs, rhs)
        self.lhs = self.nodes[0]
        self.rhs = self.nodes[1]
        self.operator = oper
        self.op_rep = op_rep

    def clone(self, nodes):
        new = self.__class__(nodes[0], nodes[1])
        return new

    def __str__(self):
        return f"({str(self.lhs)} {self.op_rep} {str(self.rhs)})"


class PredAnd(BinaryExpr):
    def __init__(self, lhs, rhs):
        super(PredAnd, self).__init__(lhs, rhs, operator.and_, 'AND')


class PredOr(BinaryExpr):
    def __init__(self, lhs, rhs):
        super(PredOr, self).__init__(lhs, rhs, operator.or_, 'OR')


class PredNot(UnaryExpr):
    def __init__(self, pred):
        super(PredNot, self).__init__(pred, operator.not_)


class PredEq(BinaryExpr):
    def __init__(self, lhs, rhs):
        super(PredEq, self).__init__(lhs, rhs, operator.eq, '==')


class PredGt(BinaryExpr):
    def __init__(self, key, value):
        super(PredGt, self).__init__(key, value, operator.gt, '>')


class PredGe(BinaryExpr):
    def __init__(self, key, value):
        super(PredGe, self).__init__(key, value, operator.ge, '>=')


class PredLe(BinaryExpr):
    def __init__(self, key, value):
        super(PredLe, self).__init__(key, value, operator.le, '<=')


class PredLt(BinaryExpr):
    def __init__(self, key, value):
        super(PredLt, self).__init__(key, value, operator.lt, '<')


class PredNe(BinaryExpr):
    def __init__(self, key, value):
        super(PredNe, self).__init__(key, value, operator.ne, '!=')


class ConstraintAdapter(object):
    """For GTC"""

    def __init__(self, key, value, oper):
        self.key = key
        self.value = value
        self.type = type(value)
        self.oper = oper

    def __repr__(self):
        return f"ConstraintAdapter(key={self.key}, value={self.value}, {self.oper}, type={self.type})"


def condition_terminal(tree):
    if tree.nodes:
        term = all(node.is_terminal() for node in tree.nodes)
        return not term
    else:
        return True


def adapter(tree):
    if tree.nodes and all(node.is_terminal() for node in tree.nodes):
        return ConstraintAdapter(tree.lhs.name, tree.rhs.value, tree.operator)


def query_expr_from_attr(attrs):
    # Create a query descriptor from a sequence for fields
    if len(attrs) == 0:
        return ConstExprTrue
    exprs = []
    # for name, dtype in descs:
    for attr in attrs:
        metadata = {'type': attr.type, 'description': attr.description}
        lhs = TagRepr(attr.name, metadata=metadata)
        rhs = Placeholder(attr.name)
        expr = (lhs == rhs)
        exprs.append(expr)
    return reduce(operator.and_, exprs)



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy as np
from astropy.modeling import Fittable1DModel, Parameter


class EnclosedGaussian(Fittable1DModel):
    '''Enclosed Gaussian model'''

    amplitude = Parameter()
    stddev = Parameter()

    @staticmethod
    def evaluate(x, amplitude, stddev):
        return amplitude * (1 - np.exp(-0.5 * (x / stddev)**2))

    @staticmethod
    def fit_deriv(x, amplitude, stddev):
        z = (x / stddev)**2
        t = np.exp(-0.5 * z)
        d_amplitude = -t + 1.0
        d_stddev = -amplitude * t * z / stddev
        return [d_amplitude, d_stddev]



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Recipes for testing"""

import numina.core
import numina.types.product as prod
import numina.types.frame as frame
import numina.core.requirements as reqs


class MasterBias(prod.DataProductMixin, frame.DataFrameType):
    pass


class MasterDark(prod.DataProductMixin, frame.DataFrameType):
    pass


class ImageTest(frame.DataFrameType):
    pass


class BiasRecipe(numina.core.BaseRecipe):
    master_bias = numina.core.Result(MasterBias)

    def __init__(self, *args, **kwds):
        super(BiasRecipe, self).__init__(*args, **kwds)
        self.simulate_error = kwds.get('simulate_error', False)


class DarkRecipe(numina.core.BaseRecipe):
    master_dark = numina.core.Result(MasterDark)


class ImageRecipe(numina.core.BaseRecipe):
    obresult = reqs.ObservationResultRequirement()
    master_bias = numina.core.Requirement(MasterBias, "Master Bias")
    master_dark = numina.core.Requirement(MasterDark, "Master Dark")
    result_image = numina.core.Result(ImageTest)


class ImageRecipeCom(numina.core.BaseRecipe):
    obresult = reqs.ObservationResultRequirement()
    accum_in = numina.core.Requirement(ImageTest, 'previous accum')
    result_image = numina.core.Result(ImageTest)
    accum = numina.core.Result(ImageTest)



#
# Copyright 2016-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import enum
from typing import TypeAlias, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self
    from .configorigin import ElementOrigin

from .generic import ComponentGeneric, InstrumentGeneric


class ElementBase:
    """Base class for objects in component collection"""

    def __init__(self, name: str, origin: "ElementOrigin | None" = None):
        self.name = name
        self.origin = origin

    def set_origin(self, origin):
        self.origin = origin

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent=None,
        properties=None,
        setup=None,
    ) -> "Self":
        obj = cls.__new__(cls)
        obj.__init__(comp_id, origin)
        return obj


class PropertiesBlock(ElementBase):
    """Class representing a properties component"""

    def __init__(
        self,
        name: str,
        properties: dict | None = None,
        origin: "ElementOrigin | None" = None,
    ):
        super().__init__(name, origin=origin)

        self.properties = {} if properties is None else properties
        self.children = {}

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent=None,
        properties=None,
        setup=None,
    ) -> "Self":
        obj = cls.__new__(cls)
        obj.__init__(comp_id, properties, origin)
        return obj


class SetupBlock(ElementBase):
    """Class representing a setup component"""

    def __init__(self, name: str, origin: "ElementOrigin | None" = None):
        super().__init__(name, origin=origin)
        self.values = {}

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent=None,
        properties=None,
        setup=None,
    ) -> "Self":
        obj = super().from_component(name, comp_id, origin, parent, properties, setup)
        if setup is not None:
            obj.values = setup.values
        return obj


class ComponentBlock(ElementBase):
    """Class representing a device component"""

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent=None,
        properties=None,
        setup=None,
    ) -> "ComponentGeneric":
        return ComponentGeneric.from_component(
            name, comp_id, origin, parent, properties, setup
        )


class InstrumentBlock(ElementBase):
    """Class representing a device component"""

    @classmethod
    def from_component(
        cls,
        name: str,
        comp_id: str,
        origin: "ElementOrigin | None" = None,
        parent=None,
        properties=None,
        setup=None,
    ) -> "InstrumentGeneric":
        return InstrumentGeneric.from_component(
            name, comp_id, origin, parent, properties, setup
        )


BlockType: TypeAlias = (
    Type[InstrumentBlock]
    | Type[ComponentBlock]
    | Type[SetupBlock]
    | Type[PropertiesBlock]
)


class ElementEnum(enum.Enum):
    ELEM_INSTRUMENT = 0
    ELEM_COMPONENT = 1
    ELEM_SETUP = 2
    ELEM_PROPERTIES = 3

    @staticmethod
    def to_class(block: "ElementEnum") -> BlockType:
        match block:
            case ElementEnum.ELEM_INSTRUMENT:
                return InstrumentBlock
            case ElementEnum.ELEM_COMPONENT:
                return ComponentBlock
            case ElementEnum.ELEM_SETUP:
                return SetupBlock
            case ElementEnum.ELEM_PROPERTIES:
                return PropertiesBlock

    @staticmethod
    def from_str(element_name: str) -> "ElementEnum":
        match element_name:
            case "instrument":
                return ElementEnum.ELEM_INSTRUMENT
            case "component":
                return ElementEnum.ELEM_COMPONENT
            case "properties":
                return ElementEnum.ELEM_PROPERTIES
            case "setup":
                return ElementEnum.ELEM_SETUP
            case _:
                raise ValueError(f"no class for {element_name}")

    @staticmethod
    def to_str(block: "ElementEnum") -> str:
        match block:
            case ElementEnum.ELEM_INSTRUMENT:
                return "instrument"
            case ElementEnum.ELEM_COMPONENT:
                return "component"
            case ElementEnum.ELEM_SETUP:
                return "setup"
            case ElementEnum.ELEM_PROPERTIES:
                return "properties"



import pkgutil
import numina.drps.drpbase
import numina.core.pipelineload as pload


def create_drp_test(names):
    """Function for creating DRP objects"""
    drps = {}
    for name in names:
        drpdata = pkgutil.get_data('numina.drps.tests', name)

        drp = pload.drp_load_data('numina', drpdata)
        drps[drp.name] = drp

    return numina.drps.drpbase.DrpGeneric(drps)



#
# Copyright 2011-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""Manage a fully qualified name for classes"""

import inspect


def fully_qualified_name(obj, sep='.'):
    """Return fully qualified name from object"""
    if inspect.isclass(obj):
        return obj.__module__ + sep + obj.__name__
    else:
        return obj.__module__ + sep + obj.__class__.__name__




from ..diskfiledal import FileFinder
from ..diskfiledal import FileFinderGTC


def test_candidates_ff(tmpdir):

    subd = ['sub1', 'sub2', 'sub3']
    for s in subd:
        tmpdir.mkdir(s)

    m = FileFinder()
    ca = m.candidates(str(tmpdir))
    assert sorted(ca) == sorted(subd)


def test_check_ff(tmpdir):

    dirname = 'sub1'
    fname1 = "hello.txt"
    fname2 = ".hidden"
    fname3 = "dir"

    newdir = tmpdir.mkdir(dirname)
    newfile1 = newdir.join(fname1)
    newfile1.write('content')

    newfile2 = newdir.join(fname2)
    newfile2.write('content')

    newdir.mkdir(fname3)

    m = FileFinder()
    # Conversion is needed only in Python < 3.6
    valid = m.check(str(newdir), fname1)
    assert valid

    invalid = m.check(str(newdir), fname2)
    assert invalid is False

    invalid = m.check(str(newdir), fname2)
    assert invalid is False


def test_candidates_ffg(tmpdir):

    subd = ['sub1', 'sub2', 'sub3']
    res = [('result.json', 1), ('sub1', 0), ('sub2', 0), ('sub3', 0)]
    for s in subd:
        tmpdir.mkdir(s)

    m = FileFinderGTC()
    ca = m.candidates(str(tmpdir))
    assert sorted(ca) == sorted(res)



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DAL base classes"""


from .daliface import DALInterface
from numina.exceptions import NoResultFound  # noqa: F401


class AbsDAL(DALInterface):
    pass


class AbsDrpDAL(DALInterface):
    def __init__(self, drps, *args, **kwargs):
        super(AbsDrpDAL, self).__init__()
        self.drps = drps

    def search_instrument_configuration_from_ob(self, obsres):
        from numina.instrument.assembly import assembly_instrument
        this_drp = self.drps.query_by_name(obsres.instrument)
        key, date_obs, keyname = this_drp.select_profile(obsres)
        ins = assembly_instrument(
            this_drp.configurations, key, date_obs, by_key=keyname)
        return ins

    def search_instrument_configuration(self, keyval, value, by_key='name'):
        from numina.instrument.assembly import assembly_instrument
        drp = self.drps.query_by_name(keyval)
        ins = assembly_instrument(
            drp.configurations, keyval, value, by_key=by_key)
        return ins

    def search_recipe(self, ins, mode, pipeline):
        drp = self.drps.query_by_name(ins)
        return drp.get_recipe_object(mode, pipeline_name=pipeline)

    def search_recipe_fqn(self, ins, mode, pipename):

        drp = self.drps.query_by_name(ins)

        this_pipeline = drp.pipelines[pipename]
        recipes = this_pipeline.recipes
        recipe_fqn = recipes[mode]
        return recipe_fqn

    def search_recipe_from_ob(self, ob, pipeline='default'):
        instrument = ob.instrument
        mode = ob.mode
        pipeline = ob.pipeline
        return self.search_recipe(instrument, mode, pipeline)



#
# Copyright 2020 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""Build a LoadableDRP from a yaml file"""


import warnings
import mimetypes


def is_fits_megara(pathname):
    """Check is any FITS"""
    if pathname.endswith('.fits'):
        return True
    else:
        return False


def is_fits_emir(pathname):
    """Check is any FITS"""
    if pathname.endswith('.fits'):
        return True
    else:
        return False


def is_json_structured(pathname):
    """Check is structured JSON"""
    import json
    # FIXME: I'm loading everything here
    with open(pathname) as fd:
        state = json.load(fd)

    if 'type_fqn' in state:
        return True
    else:
        return False


class DataLoaders:
    def __init__(self):
        self._loaders = []

    def register(self, mtype, is_func=None, priority=20):

        if is_func is None:
            def is_func(p): return True  # noqa: E731

        def wrapper(func):
            self._loaders.append((priority, mtype, is_func, func))
            self._loaders.sort()
            return func

        return wrapper

    def dispatch(self, pathname):

        mmtype, enc = mimetypes.guess_type(pathname)
        # This is ordered by priority
        for priority, mtype, is_func, func in self._loaders:
            if (mmtype == mtype) and is_func(pathname):
                return func(pathname)
        else:
            raise TypeError(f'nothing handles {pathname}')

    def __call__(self, pathname):
        return self.dispatch(pathname)


class DataChecker:
    def __init__(self):
        self._loaders = {}

    def register(self, instrument_name):

        def wrapper(func):
            self._loaders[instrument_name] = func
            return func

        return wrapper

    def dispatch(self, instrument, hdulist, astype=None, level=None):
        try:
            func = self._loaders[instrument]
        except KeyError:
            # No function registered
            warnings.warn(f"no function for {instrument}")
            return
        return func(hdulist, astype=astype, level=level)

    def __call__(self, instrument, hdulist, astype=None, level=None):
        return self.dispatch(instrument, hdulist, astype=astype, level=level)


if __name__ == '__main__':
    load = DataLoaders()

    @load.register('image/fits', priority=20)
    def load_fits_0(pathname):
        import astropy.io.fits as fits
        return fits.open(pathname)

    @load.register('image/fits', is_fits_megara, priority=19)
    def load_fits_1(pathname):
        import astropy.io.fits as fits
        return fits.open(pathname)

    @load.register('image/fits', is_fits_emir, priority=5)
    def load_fits_2(pathname):
        import astropy.io.fits as fits
        return fits.open(pathname)

    @load.register('application/json', priority=20)
    def load_json(pathname):
        import json
        with open(pathname) as fd:
            return json.load(fd)

    @load.register('application/json', is_json_structured, priority=5)
    def load_json(pathname):  # noqa: F811
        import json
        from numina.util.objimport import import_object

        with open(pathname) as fd:
            data = json.load(fd)
        type_fqn = data['type_fqn']
        cls = import_object(type_fqn)
        obj = cls.__new__(cls)
        obj.__setstate__(data)
        return obj

    fname = '/home/spr/devel/guaix/megaradrp/problem_marisa/data/master_traces.json'
    print(is_json_structured(fname))
    value = load.dispatch(fname)
    print(value)



#
# Copyright 2008-2016 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


def default_dialect_info(obj):
    key = obj.__module__ + '.' + obj.__class__.__name__
    result = {'base': {'fqn': key, 'python': obj.internal_type}}
    return result


dialect_info = default_dialect_info



#
# Copyright 2015-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

"""DAL for dictionary-based database of products."""

import os
import logging
import json
from itertools import chain

import yaml

import numina.store
import numina.store.gtc.load as gtcload
from numina.core.oresult import oblock_from_dict
from numina.exceptions import NoResultFound
from numina.instrument.assembly import assembly_instrument
from numina.util.context import working_directory

from .absdal import AbsDrpDAL
from .stored import ObservingBlock
from .stored import StoredProduct, StoredParameter, StoredResult
from .diskfiledal import build_product_path
from .utils import tags_are_valid


_logger = logging.getLogger(__name__)


class BaseDictDAL(AbsDrpDAL):
    """A dictionary based DAL"""

    _RESERVED_MODE_NAMES = ['nulo', 'container', 'root', 'raiz']

    def __init__(self, drps, ob_table, prod_table, req_table, extra_data=None, components=None):
        super(BaseDictDAL, self).__init__(drps)
        # Check that the structure of the base is correct
        self.ob_table = ob_table
        self.prod_table = prod_table
        self.req_table = req_table
        self.extra_data = extra_data if extra_data else {}
        self.components = components if components else {}

    def search_oblock_from_id(self, obsid):
        try:
            ob = self.ob_table[obsid]
            return ObservingBlock(**ob)
        except KeyError:
            raise NoResultFound(f"oblock with id {obsid} not found")

    def search_prod_obsid(self, ins, obsid, pipeline):
        """Returns the first coincidence..."""
        ins_prod = self.prod_table.get(ins, {})
        for profile, prof_prod in ins_prod.items():
            for prod in prof_prod:
                if prod['ob'] == obsid:
                    # We have found the result, no more checks
                    return StoredProduct(**prod)
        else:
            raise NoResultFound(f'result for ob {obsid} not found')

    def search_prod_req_tags(self, req, ins, profile, tags, pipeline):
        if req.dest in self.extra_data:
            val = self.extra_data[req.dest]
            content = numina.store.load(req.type, val)
            return StoredProduct(id=0, tags={}, content=content)
        else:
            return self.search_prod_type_tags(req.type, ins, profile, tags, pipeline)

    def search_prod_type_tags(self, tipo, ins, profile, tags, pipeline):
        """Returns the first coincidence..."""

        drp = self.drps.query_by_name(ins)
        label = drp.product_label(tipo)

        # Strip () is present
        if label.endswith("()"):
            label_alt = label[:-2]
        else:
            label_alt = label

        # search results of these OBs
        ins_table = self.prod_table.get(ins, {})
        ptable = ins_table.get(profile, [])
        for prod in ptable:
            pk = prod['type']
            pt = prod['tags']
            if ((pk == label) or (pk == label_alt)) and tags_are_valid(pt, tags):
                # this is a valid product
                # We have found the result, no more checks
                # Make a copy
                rprod = dict(prod)
                rprod['content'] = numina.store.load(tipo, prod['content'])
                return StoredProduct(**rprod)
        else:
            msg = f'type {tipo} compatible with tags {tags!r} not found'
            raise NoResultFound(msg)

    def search_param_req(self, req, instrument, profile, mode, pipeline):
        req_table_ins = self.req_table.get(instrument, {})
        req_table_ins_p = req_table_ins.get(profile, {})
        req_table_insi_pipe = req_table_ins_p.get(pipeline, {})
        mode_keys = req_table_insi_pipe.get(mode, {})
        if req.dest in self.extra_data:
            value = self.extra_data[req.dest]
            content = StoredParameter(value)
            return content
        elif req.dest in mode_keys:
            value = mode_keys[req.dest]
            content = StoredParameter(value)
            return content
        else:
            raise NoResultFound(
                f"No parameters for {mode} mode, pipeline {pipeline}")

    def search_param_req_tags(self, req, instrument, profile, mode, tags, pipeline):
        req_table_ins = self.req_table.get(instrument, {})
        req_table_ins_p = req_table_ins.get(profile, {})
        req_table_insi_pipe = req_table_ins_p.get(pipeline, {})
        mode_list = req_table_insi_pipe.get(mode, [])
        if req.dest in self.extra_data:
            value = self.extra_data[req.dest]
            content = StoredParameter(value)
            return content
        else:
            for prod in mode_list:
                pn = prod['name']
                pt = prod['tags']
                if pn == req.dest and tags_are_valid(pt, tags):
                    # We have found the result, no more checks
                    value = numina.store.load(req.type, prod['content'])
                    content = StoredParameter(value)
                    return content
            else:
                msg = f'name {req.dest} compatible with tags {tags!r} not found'
                raise NoResultFound(msg)

    def oblock_from_id(self, obsid):

        este = self.ob_table[obsid]
        oblock = oblock_from_dict(este)

        return oblock

    def obsres_from_oblock_id(self, obsid, as_mode=None, configuration=None):
        """"
        Override instrument configuration if configuration is not None
        """

        oblock = self.oblock_from_id(obsid)

        return self.obsres_from_oblock(oblock, as_mode)

    def obsres_from_oblock(self, oblock, as_mode=None):

        from numina.core.oresult import ObservationResult

        # Internal copy
        obsres = ObservationResult.__new__(ObservationResult)
        obsres.__dict__ = oblock.__dict__

        obsres.mode = as_mode or obsres.mode
        _logger.debug(
            "obsres_from_oblock id='%s', mode='%s' START", obsres.id, obsres.mode)

        try:
            this_drp = self.drps.query_by_name(obsres.instrument)
        except KeyError:
            raise ValueError(f'no DRP for instrument {obsres.instrument}')

        # Reserved names
        if obsres.mode in self._RESERVED_MODE_NAMES:
            selected_mode = None  # null mode
        else:
            selected_mode = this_drp.modes[obsres.mode]

        if selected_mode:
            # This is used if we pass an option here or
            # if mode.build_ob_options is defined in the mode class
            # seems useless
            obsres = selected_mode.build_ob(obsres, self)
            # Not needed, all the information is obtained from
            # the requirements
            obsres = selected_mode.tag_ob(obsres)

        _logger.debug('assembly instrument model, auto detection')
        key, date_obs, keyname = this_drp.select_profile(obsres)
        obsres.configuration = self.assembly_instrument(key, date_obs, keyname)
        obsres.profile = str(obsres.configuration.origin.uuid)
        _logger.debug(f'instrument profile is {obsres.profile}')

        auto_configure = True
        sample_frame = obsres.get_sample_frame()
        if auto_configure and sample_frame is not None:
            _logger.debug('configuring instrument model with image from obsres')
            obsres.configuration.configure_with_image(sample_frame.open())
        else:
            _logger.debug('no configuring instrument model')
        return obsres

    def assembly_instrument(self, keyval, date, by_key='name'):
        return assembly_instrument(self.components, keyval, date, by_key=by_key)

    def search_result_id(self, node_id, tipo, field):
        cobsres = self.obsres_from_oblock_id(node_id)

        rdir = resultsdir_default(self.basedir, node_id)
        # FIXME: hardcoded
        # taskfile = os.path.join(rdir, 'task.yaml')
        resfile = os.path.join(rdir, 'result.yaml')
        result_contents = yaml.safe_load(open(resfile))
        # task_contents = yaml.safe_load(open(taskfile))

        try:
            field_file = result_contents[field]
            st = StoredProduct(
                id=node_id,
                content=numina.store.load(
                    tipo, os.path.join(rdir, field_file)),
                tags={}
            )
            return st
        except KeyError as err:
            msg = f"field '{field}' not found in result of mode '{cobsres.mode}' id={node_id}"
            raise NoResultFound(msg) from err

    def search_product(self, name, tipo, obsres, options=None):
        # returns StoredProduct
        ins = obsres.instrument
        profile = obsres.profile
        tags = obsres.tags
        pipeline = obsres.pipeline
        if name in self.extra_data:
            val = self.extra_data[name]
            content = numina.store.load(tipo, val)
            return StoredProduct(id=0, tags={}, content=content)
        else:
            return self.search_prod_type_tags(tipo, ins, profile, tags, pipeline)

    def search_parameter(self, name, tipo, obsres, options=None):
        # returns StoredProduct
        instrument = obsres.instrument
        profile = obsres.profile
        mode = obsres.mode
        tags = obsres.tags
        pipeline = obsres.pipeline

        req_table_ins = self.req_table.get(instrument, {})
        req_table_ins_p = req_table_ins.get(profile, {})
        req_table_insi_pipe = req_table_ins_p.get(pipeline, {})

        mode_list = req_table_insi_pipe.get(mode, [])
        if name in self.extra_data:
            value = self.extra_data[name]
            content = StoredParameter(value)
            return content
        else:
            for prod in mode_list:
                pn = prod['name']
                pt = prod['tags']
                if pn == name and tags_are_valid(pt, tags):
                    # We have found the result, no more checks
                    value = numina.store.load(tipo, prod['content'])
                    content = StoredParameter(value)
                    return content
            else:
                msg = f'name {name} compatible with tags {tags!r} not found'
                raise NoResultFound(msg)

    def search_result_relative(self, name, tipo, obsres, result_desc, options=None):
        # mode field node could go together...
        return []


class DictDAL(BaseDictDAL):
    def __init__(self, drps, base):

        # Check that the structure of 'base' is correct
        super(DictDAL, self).__init__(
            drps,
            base['oblocks'],
            base['products'],
            base['parameters']
        )


class Dict2DAL(BaseDictDAL):
    def __init__(self, drps, obtable, base, extra_data=None, components=None):

        prod_table = base.get('products', {})

        if 'parameters' in base:
            req_table = base['parameters']
        else:
            req_table = base.get('requirements', {})

        super(Dict2DAL, self).__init__(
            drps, obtable, prod_table, req_table,
            extra_data, components=components
        )

    def new_task_id(self, request, request_params):
        if request == 'reduce':
            return request_params.get('oblock_id', 1)
        return 1

    def new_task(self, request, request_params):

        from numina import __version__
        from numina.user.helpers import ProcessingTask

        newidx = self.new_task_id(request, request_params)
        _logger.debug('create task=%s', newidx)
        task = ProcessingTask()
        task.id = newidx
        task.request = request
        task.request_params = request_params
        task.request_runinfo['runner'] = 'numina'
        task.request_runinfo['runner_version'] = __version__
        return task

    def update_task(self, task):
        pass

    def update_result(self, task, serialized, filename):
        pass

    def dump(self, fp):
        state = self.dump_data()
        yaml.dump(state, fp)
        # yaml.dump(state, fp, default_flow_style=False)

        with open('control_dump.json', 'w') as fp:
            json.dump(state, fp, indent=2)

    def dump_data(self):
        state = {}
        state['version'] = 1
        state['products'] = self.prod_table
        state['requirements'] = self.req_table
        state['oblocks'] = self.ob_table
        return state


def workdir_default(basedir, obsid):
    # FIXME: hardcoded
    workdir = os.path.join(basedir, f'obsid{obsid}_work')
    workdir = os.path.abspath(workdir)
    return workdir


def resultsdir_default(basedir, obsid):
    # FIXME: hardcoded
    resultsdir = os.path.join(basedir, f'obsid{obsid}_results')
    resultsdir = os.path.abspath(resultsdir)
    return resultsdir


class BaseHybridDAL(Dict2DAL):

    def __init__(self, drps, obtable, base, extra_data=None, basedir=None, components=None):

        self.rootdir = base.get("rootdir", "")
        self.ob_ids = []

        if basedir is None:
            self.basedir = os.getcwd()
        else:
            self.basedir = basedir

        super(BaseHybridDAL, self).__init__(
            drps, obtable, base, extra_data=extra_data, components=components
        )

    def add_obs(self, obtable):
        obdict = {}
        for ob in obtable:
            obid = ob['id']
            if obid not in self.ob_ids:

                self.ob_ids.append(obid)
                obdict[obid] = ob
            else:
                _logger.warning('oblock_id=%s is already in table', obid)
        # Update parents
        for ob in obdict.values():
            children = ob.get('children', [])
            for ch in children:
                obdict[ch]['parent'] = ob['id']

        self.ob_table.update(obdict)

    def search_product(self, name, tipo, obsres, options=None):
        if name in self.extra_data:
            val = self.extra_data[name]
            content = numina.store.load(tipo, val)
            return StoredProduct(id=0, tags={}, content=content)
        else:
            return self._search_prod_table(name, tipo, obsres)

    def _search_prod_table(self, name, tipo, obsres):
        raise NotImplementedError

    def search_result(self, name, tipo, obsres, resultid=None):

        if resultid is None:
            for g in chain([tipo.name()], tipo.generators()):
                if g in obsres.results:
                    resultid = obsres.results[g]
                    break
            else:
                raise NoResultFound("resultid not found")
        prod = self._search_result(name, tipo, obsres, resultid)
        return prod

    def _search_result(self, name, tipo, obsres, resultid):
        """Returns the first coincidence..."""

        instrument = obsres.instrument
        profile = obsres.profile

        drp = self.drps.query_by_name(instrument)
        # label = drp.product_label(tipo)

        # search results of these OBs
        ins_tab = self.prod_table.get(instrument, {})
        ptable = ins_tab.get(profile, [])
        for prod in ptable:
            pid = prod['id']
            if pid == resultid:
                # this is a valid product
                # We have found the result, no more checks
                # Make a copy
                rprod = dict(prod)

                if 'content' in prod:
                    path = prod['content']
                else:
                    # Build path
                    path = build_product_path(
                        drp, self.rootdir, profile, name, tipo, obsres)
                _logger.debug("searching product in path: %s", path)
                rprod['content'] = self.product_loader(tipo, name, path)
                return StoredProduct(**rprod)
        else:
            msg = f'result with id {resultid} not found'
            raise NoResultFound(msg)

    def product_loader(self, tipo, name, path):
        path, kind = path
        if kind == 0:
            return numina.store.load(tipo, path)
        else:
            # GTC load
            with open(path) as fd:
                data = json.load(fd)
                inter = gtcload.build_result(data)
                elem = inter['elements']
                return elem[name]

    def search_previous_obsres(self, obsres, node=None):

        if node is None:
            node = 'prev'

        if node == 'prev-rel':
            # Compute nodes relative to parent
            # unless parent is None, then is equal to prev
            parent_id = obsres.parent
            if parent_id is not None:
                cobsres = self.obsres_from_oblock_id(parent_id)
                subset_ids = cobsres.children
                idx = subset_ids.index(obsres.id)
                return reversed(subset_ids[:idx])
            else:
                return self.search_previous_obsid_all(obsres.id)
        else:
            return self.search_previous_obsid_all(obsres.id)

    def search_previous_obsid_all(self, obsid):
        idx = self.ob_ids.index(obsid)
        return reversed(self.ob_ids[:idx])

    def search_result_id(self, node_id, tipo, field, mode=None):
        raise NotImplementedError

    def search_result_relative(self, name, tipo, obsres, result_desc, options=None):

        _logger.debug('search relative result for %s', name)

        # result_type = DataFrameType()
        result_mode = result_desc.mode
        result_field = result_desc.attr
        result_node = result_desc.node

        ignore_fail = result_desc.ignore_fail

        if result_node == 'children':
            # Results are multiple
            # one per children
            _logger.debug('search children nodes of %s', obsres.id)
            results = []
            for c in obsres.children:
                try:
                    st = self.search_result_id(c, tipo, result_field)
                    results.append(st)
                except NoResultFound:
                    if not ignore_fail:
                        raise

            return results
        elif result_node == 'prev' or result_node == 'prev-rel':
            _logger.debug('search previous nodes of %s', obsres.id)

            # obtain previous nodes
            for previd in self.search_previous_obsres(obsres, node=result_node):
                # print('searching in node', previd)
                try:
                    st = self.search_result_id(
                        previd, tipo, result_field, mode=result_mode)
                    return st
                except NoResultFound:
                    pass

            else:
                raise NoResultFound('value not found in any node')
        elif result_node == 'last':
            _logger.debug('search last node of %s', result_mode)

            try:
                st = self.search_result_last(name, tipo, result_desc)
                return st
            except NoResultFound:
                pass
        else:
            msg = f'unknown node type {result_node}'
            raise TypeError(msg)

    def search_result_last(self, name, tipo, result_desc):
        # FIXME: Implement
        raise NoResultFound

    def build_product_path(self, drp, conf, name, tipo, obsres):
        path = build_product_path(drp, self.rootdir, conf, name, tipo, obsres)
        return path

    def search_session_ids(self):
        for obs_id in self.ob_ids:
            obdict = self.ob_table[obs_id]
            enabled = obdict.get('enabled', True)
            if (not enabled) or obdict['mode'] in self._RESERVED_MODE_NAMES:
                # ignore these OBs
                continue

            yield obs_id

    def dump_data(self):
        state = super(BaseHybridDAL, self).dump_data()
        state['rootdir'] = self.rootdir
        state['ob_ids'] = self.ob_ids
        return state


class HybridDAL(BaseHybridDAL):
    """A DAL that can read files from directory structure"""

    def __init__(self, drps, obtable, base, extra_data=None, components=None, basedir=None):

        temp_ob_ids = []
        # Preprocessing
        obdict = {}
        for ob in obtable:
            obid = ob['id']
            temp_ob_ids.append(obid)
            obdict[obid] = ob

        # Update parents
        for ob in obdict.values():
            children = ob.get('children', [])
            for ch in children:
                obdict[ch]['parent'] = ob['id']

        super(HybridDAL, self).__init__(
            drps, obdict, base,
            extra_data=extra_data,
            basedir=basedir,
            components=components
        )
        # This field does not exist until super is called
        self.ob_ids = temp_ob_ids

    def _search_prod_table(self, name, tipo, obsres):
        """Returns the first coincidence..."""

        instrument = obsres.instrument

        profile = obsres.profile

        drp = self.drps.query_by_name(instrument)
        label = drp.product_label(tipo)
        # Strip () is present
        if label.endswith("()"):
            label_alt = label[:-2]
        else:
            label_alt = label

        # search results of these OBs
        ins_table = self.prod_table.get(instrument, {})
        ptable = ins_table.get(profile, [])
        for prod in ptable:
            pk = prod['type']
            pt = prod['tags']
            if ((pk == label) or (pk == label_alt)) and tags_are_valid(pt, obsres.tags):
                # this is a valid product
                # We have found the result, no more checks
                # Make a copy
                rprod = dict(prod)

                if 'content' in prod:
                    path = prod['content']
                else:
                    # Build path
                    path = build_product_path(
                        drp, self.rootdir, profile, name, tipo, obsres)
                _logger.debug("path is %s", path)
                rprod['content'] = numina.store.load(tipo, path)
                return StoredProduct(**rprod)
        else:
            # Not in table, try file directly
            _logger.debug("%s not in table, try file directly", tipo)
            path = self.build_product_path(drp, profile, name, tipo, obsres)
            _logger.debug("path is %s", path)
            content = self.product_loader(tipo, name, path)
            return StoredProduct(id=0, content=content, tags=obsres.tags)

    def search_result_id(self, node_id, tipo, field, mode=None):
        cobsres = self.obsres_from_oblock_id(node_id)

        if mode is not None:
            # mode must match
            if cobsres.mode != mode:
                msg = f"requested mode '{mode}' and obsmode '{cobsres.mode}' do not match"
                raise NoResultFound(msg)

        try:
            directory = resultsdir_default(self.basedir, node_id)

            # change directory to open result file
            with working_directory(os.path.join(self.basedir, directory)):

                # Try to open both
                filename_yaml = 'result.yaml'
                filename_json = 'result.json'
                if os.path.exists(filename_yaml):
                    with open(filename_yaml) as fd:
                        result_data = yaml.safe_load(fd)
                elif os.path.exists(filename_json):
                    with open(filename_json) as fd:
                        result_data = json.load(fd)
                else:
                    raise ValueError(
                        f'result.yaml or result.json not found in {directory}')

                stored_result = StoredResult.load_data(result_data)

                try:
                    content = getattr(stored_result, field)
                except AttributeError:
                    raise NoResultFound(f'no field {field} found in result')

                st = StoredProduct(
                    id=node_id,
                    content=content,
                    tags={}
                )
                return st
        except KeyError as err:
            msg = f"field '{field}' not found in result of mode '{cobsres.mode}' id={node_id}"
            raise NoResultFound(msg) from err



#
# Copyright 2017-2018 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


import json
import datetime

import numpy

import numina.types.qc


class ExtEncoder(json.JSONEncoder):
    """"Encode numpy.floats and numpy.integer"""

    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        elif isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        elif isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        elif isinstance(obj, numina.types.qc.QC):
            return obj.name
        else:
            return super(ExtEncoder, self).default(obj)



#
# Copyright 2024-2025 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import astropy.constants as constants
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np

from .raise_valueerror import raise_ValueError


def simulate_spectrum(wave, flux, flux_type,
                      nphotons, wavelength_sampling,
                      rng, wmin, wmax, convolve_sigma_km_s,
                      nbins_histo, plots, plot_title, verbose):
    """Simulate spectrum defined by tabulated wave and flux data.

    Parameters
    ----------
    wave : `~astropy.units.Quantity`
        Numpy array (with astropy units) containing the tabulated
        wavelength.
    flux : array_like
        Array-like object containing the tabulated flux.
    flux_type : str
        Relative flux unit. Valid options are:
        - flam: proportional to erg s^-1 cm^-2 A^-1
        - photlam: proportional to photon s^-1 cm^-2 A^-1
    nphotons : int
        Number of photons to be simulated
    wavelength_sampling : str
        Method to sample the wavelength values. Two options are valid:
        - 'random': the wavelengt of each photon is randomly determined
          using the spectrum shape as the density probability function.
        - 'fixed': the wavelength of each photon is exactly determined
          using the spectrum shape as the density probability function.
    rng : `~numpy.random._generator.Generator`
        Random number generator.
    wmin : `~astropy.units.Quantity`
        Minimum wavelength to be considered.
    wmax : `~astropy.units.Quantity`
        Maximum wavelength to be considered.
    convolve_sigma_km_s : `~astropy.units.Quantity`
        Gaussian broadening (sigma) in km/s to be applied.
    nbins_histo : int
        Number of bins for histogram plot.
    plots : bool
        If True, plot input and output results.
    plot_title : str or None
        Plot title. Used only when 'plots' is True.
    verbose : bool
        If True, display additional information.

    Returns
    -------
    simulated_wave : `~astropy.units.Quantity`
        Wavelength of simulated photons.
    """

    flux = np.asarray(flux)
    if len(wave) != len(flux):
        raise_ValueError(f"Incompatible array length: 'wave' ({len(wave)}), 'flux' ({len(flux)})")

    if np.any(flux < 0):
        raise_ValueError(f'Negative flux values cannot be handled')

    if flux_type.lower() not in ['flam', 'photlam']:
        raise_ValueError(f"Flux type: {flux_type} is not any of the valid values: 'flam', 'photlam'")

    if not isinstance(wave, u.Quantity):
        raise_ValueError(f"Object {wave=} is not a Quantity instance")
    wave_unit = wave.unit
    if not wave_unit.is_equivalent(u.m):
        raise_ValueError(f"Unexpected unit for 'wave': {wave_unit}")

    # lower wavelength limit
    if wmin is not None:
        if not isinstance(wmin, u.Quantity):
            raise_ValueError(f"Object {wmin=} is not a Quantity instance")
        if not wmin.unit.is_equivalent(u.m):
            raise_ValueError(f"Unexpected unit for 'wmin': {wmin}")
        wmin = wmin.to(wave_unit)
        lower_index = np.searchsorted(wave.value, wmin.value, side='left')
    else:
        lower_index = 0

    # upper wavelength limit
    if wmax is not None:
        if not isinstance(wmax, u.Quantity):
            raise_ValueError(f"Object {wmax=} is not a Quantity instance")
        if not wmax.unit.is_equivalent(u.m):
            raise_ValueError(f"Unexpected unit for 'wmax': {wmin}")
        wmax = wmax.to(wave_unit)
        upper_index = np.searchsorted(wave.value, wmax.value, side='right')
    else:
        upper_index = len(wave)

    if lower_index == upper_index:
        if plot_title is not None:
            print(f'Working with data from: {plot_title}')
        print(f'Tabulated wavelength range: {wave[0]} - {wave[-1]}')
        print(f'Requested wavelength range: {wmin} - {wmax}')
        raise_ValueError('Wavelength ranges without intersection')

    if not isinstance(convolve_sigma_km_s, u.Quantity):
        raise_ValueError(f'Object {convolve_sigma_km_s=} is not a Quantity instance')
    if convolve_sigma_km_s.unit != u.km / u.s:
        raise_ValueError(f'Unexpected unit for {convolve_sigma_km_s}')
    if convolve_sigma_km_s.value < 0:
        raise_ValueError(f'Unexpected negative value for {convolve_sigma_km_s}')

    if plots:
        fig, ax = plt.subplots()
        ax.plot(wave.value, flux, '-')
        if wmin is not None:
            ax.axvline(wmin.value, linestyle='--', color='gray')
        if wmax is not None:
            ax.axvline(wmax.value, linestyle='--', color='gray')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Flux (arbitrary units)')
        if plot_title is not None:
            ax.set_title(plot_title)
        plt.tight_layout()
        plt.show()

    wave = wave[lower_index:upper_index]
    flux = flux[lower_index:upper_index]

    # convert FLAM to PHOTLAM
    if flux_type.lower() == 'flam':
        if verbose:
            print('Converting FLAM to PHOTLAM')
        flux_conversion = wave.to(u.m) / (constants.h * constants.c)
        flux *= flux_conversion.value

    wmin_eff = wave[0]
    wmax_eff = wave[-1]

    # normalized cumulative area
    # (area under the polygons defined by the tabulated data)
    cumulative_area = np.concatenate((
        [0],
        np.cumsum((flux[:-1] + flux[1:])/2 * (wave[1:] - wave[:-1]))
    ))
    normalized_cumulative_area = cumulative_area / cumulative_area[-1]

    if plots:
        fig, ax = plt.subplots()
        ax.plot(wave.value, normalized_cumulative_area, '.')
        ax.axvline(wmin_eff.value, linestyle='--', color='gray')
        ax.axvline(wmax_eff.value, linestyle='--', color='gray')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Normalized cumulative area')
        if plot_title is not None:
            ax.set_title(plot_title)
        plt.tight_layout()
        plt.show()

    if wavelength_sampling == 'random':
        # samples following a uniform distribution
        unisamples = rng.uniform(low=0, high=1, size=nphotons)
    elif wavelength_sampling == 'fixed':
        # constant sampling of distribution function
        unisamples = np.linspace(0, 1, num=nphotons+1)  # generate extra photon
        unisamples = unisamples[:-1]  # remove last photon
    else:
        unisamples = None  # avoid PyCharm warning
        raise_ValueError(f'Unexpected {wavelength_sampling=}')

    simulated_wave = np.interp(x=unisamples, xp=normalized_cumulative_area, fp=wave.value)

    # apply Gaussian broadening
    if convolve_sigma_km_s.value > 0:
        if verbose:
            print(f'Applying {convolve_sigma_km_s=}')
        sigma_wave = convolve_sigma_km_s / constants.c.to(u.km / u.s) * simulated_wave
        simulated_wave = rng.normal(loc=simulated_wave, scale=sigma_wave)

    # add units
    simulated_wave *= wave_unit

    if plots:
        fig, ax = plt.subplots()
        hist_sim, bin_edges_sim = np.histogram(simulated_wave.value, bins=nbins_histo)
        xhist_sim = (bin_edges_sim[:-1] + bin_edges_sim[1:]) / 2
        fscale = np.median(hist_sim / np.interp(x=xhist_sim, xp=wave.value, fp=flux))
        ax.plot(wave.value, flux*fscale, 'k-', linewidth=1, label='rescaled input spectrum')
        hist_dum = np.diff(np.interp(x=bin_edges_sim, xp=wave.value, fp=normalized_cumulative_area)) * nphotons
        ax.plot(xhist_sim, hist_dum, '-', linewidth=3, label='binned input spectrum')
        ax.plot(xhist_sim, hist_sim, '-', linewidth=1, label='binned simulated spectrum')
        ax.axvline(wmin_eff.value, linestyle='--', color='gray')
        ax.axvline(wmax_eff.value, linestyle='--', color='gray')
        ax.set_xlabel(f'Wavelength ({wave_unit})')
        ax.set_ylabel('Number of simulated photons')
        if plot_title is not None:
            ax.set_title(plot_title)
        ax.legend()
        plt.tight_layout()
        plt.show()

    return simulated_wave



#
# Copyright 2008-2021 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#


"""
Recipe inputs and outputs
"""

import uuid
import logging

from .metaclass import RecipeInputType, RecipeResultType
import numina.store.dump
import numina.types.qc


_logger = logging.getLogger(__name__)


class RecipeInOut(object):

    def __init__(self, *args, **kwds):
        super(RecipeInOut, self).__init__()
        # Used to hold set values
        # Use this to avoid infinite recursion
        super(RecipeInOut, self).__setattr__('_numina_desc_val', {})
        # instead of this
        # self._numina_desc_val = {}
        all_msg_errors = []

        # memorize aliases
        super(RecipeInOut, self).__setattr__('_aliases', {})

        for key, req in self.stored().items():
            if req.alias:
                self._aliases[req.alias] = req

        for key, val in kwds.items():
            try:
                setattr(self, key, kwds[key])
            except (ValueError, TypeError) as err:
                all_msg_errors.append(err.args[0])

        self._finalize(all_msg_errors)

    def __repr__(self):
        sclass = type(self).__name__
        full = []
        for key, val in self.stored().items():
            full.append(f'{key}={val!r}')
        return f"{sclass}({', '.join(full)})"

    def __getattr__(self, item):
        # This method might be called before _aliases is initialized
        if item in self.__dict__.get('_aliases', {}):
            ref = self.__dict__['_aliases'][item]
            return getattr(self, ref.dest)
        else:
            msg = f"'{self.__class__.__name__}' object has no attribute '{item}'"
            raise AttributeError(msg)

    def __setattr__(self, item, value):
        # This method might be called before _aliases is initialized
        if item in self.__dict__.get('_aliases', {}):
            ref = self.__dict__['_aliases'][item]
            return setattr(self, ref.dest, value)
        else:
            super(RecipeInOut, self).__setattr__(item, value)

    def _finalize(self, all_msg_errors=None):
        """Access all the instance descriptors

        This wil trigger an exception if a required
        parameter is not set
        """
        if all_msg_errors is None:
            all_msg_errors = []

        for key in self.stored():
            try:
                getattr(self, key)
            except (ValueError, TypeError) as err:
                all_msg_errors.append(err.args[0])

        # Raises a list of all the missing entries
        if all_msg_errors:
            raise ValueError(all_msg_errors)

    def attrs(self):
        return self._numina_desc_val

    @classmethod
    def stored(cls):
        return cls.__numina_stored__

    def validate(self):
        """Validate myself."""

        for key, req in self.stored().items():
            val = getattr(self, key)
            _logger.info('validate %s with a value of %s', req, val)
            try:
                req.validate(val)
                _logger.info('validation passed')
            except Exception as error:
                _logger.warning('validation failed with error %s', error)

        # Run checks defined in __checkers__
        self._run_checks()

    def _run_checks(self):
        checkers = getattr(self, '__checkers__', [])

        for check in checkers:
            check.check(self)

    @classmethod
    def tag_names(cls):
        qfields = set()
        for key, req in cls.stored().items():
            tag_n = req.tag_names()
            qfields.update(tag_n)
        return qfields


class RecipeInput(RecipeInOut, metaclass=RecipeInputType):
    """RecipeInput base class"""
    pass


class RecipeResultBase(RecipeInOut, metaclass=RecipeResultType):
    """The result of a Recipe."""

    def store_to(self, where):

        saveres = dict(values={})
        saveres_v = saveres['values']
        for key, prod in self.stored().items():
            val = getattr(self, key)
            saveres_v[key] = numina.store.dump(prod.type, val, prod.dest)

        return saveres


class RecipeResult(RecipeResultBase):

    def __init__(self, *args, **kwds):

        # Extract QC if available
        self.qc = numina.types.qc.QC.UNKNOWN
        self.uuid = uuid.uuid1()

        # qc is not passed further
        if 'qc' in kwds:
            self.qc = kwds['qc']
            del kwds['qc']

        super(RecipeResult, self).__init__(*args, **kwds)

    def store_to(self, where):
        saveres = super(RecipeResult, self).store_to(where)

        saveres['qc'] = self.qc.name
        saveres['uuid'] = str(self.uuid)
        return saveres

    def time_it(self, time1, time2):
        import numina.types.dataframe as dataframe
        values = self.attrs()
        for k, spec in self.stored().items():
            value = values[k]
            # Store for Images..
            if isinstance(value, dataframe.DataFrame):
                hdul = value.open()
                self.add_computation_time(hdul, time1, time2)

    def add_computation_time(self, img, time1, time2):
        img[0].header['NUMUTC1'] = time1.isoformat()
        img[0].header['NUMUTC2'] = time2.isoformat()
        return img


class define_result(object):
    """Recipe decorator."""

    def __init__(self, resultClass):
        if not issubclass(resultClass, RecipeResult):
            msg = f'{resultClass!r} does not derive from RecipeResult'
            raise TypeError(msg)
        self.klass = resultClass

    def __call__(self, klass):
        klass.RecipeResult = self.klass
        return klass


class define_input(object):
    """Recipe decorator."""

    def __init__(self, input_class):
        if not issubclass(input_class, RecipeInput):
            msg = f'{input_class!r} does not derive from RecipeInput'
            raise TypeError(msg)
        self.klass = input_class

    def __call__(self, klass):
        klass.RecipeInput = self.klass
        return klass


define_requirements = define_input



#
# Copyright 2008-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import numpy

from numina.array.cosmetics import cosmetics, ccdmask


def test_cosmetics_flat2_None():
    flat1 = numpy.ones((20, 20))
    mascara = cosmetics(flat1)
    assert not bool(mascara.all())


def test_cosmetics_no_mask():
    flat1 = numpy.ones((20, 20))
    flat2 = flat1
    mascara = cosmetics(flat1, flat2)
    assert not mascara.all()


def test_cosmetics_mask():
    flat1 = numpy.ones((2, 2))
    flat2 = flat1
    mask = numpy.zeros((2, 2), dtype='int')
    mascara = cosmetics(flat1, flat2, mask)
    assert not mascara.all()


def test_cosmetics_mask_2():
    size = 200
    flat1 = numpy.ones((size, size))
    flat1[0:100, 0] = 0
    flat2 = flat1
    mask = numpy.zeros((size, size), dtype='int')
    mask[0:10, 0] = 1
    mascara = cosmetics(flat1, flat2, mask)
    expected_mask = numpy.zeros((size, size), dtype='int')
    expected_mask[0:100, 0] = 1
    expected_mask = expected_mask.astype('bool')
    assert mascara.all() == expected_mask.all()


def test_ccdmask_flat2_None():
    flat1 = numpy.ones((20, 20))
    bpm2 = ccdmask(flat1)
    assert not bpm2[1].all()


def test_ccdmask_no_mask():
    flat1 = numpy.ones((2000, 2000))
    flat1[0:100, 0] = 0
    flat2 = flat1
    bpm2 = ccdmask(flat1, flat2, mode='full')
    assert bpm2[1].sum() == 100


def test_ccdmask_empty_mask():
    flat1 = numpy.ones((2000, 2000))
    flat1[0:100, 0] = 0
    flat2 = flat1
    mask = numpy.zeros((2000, 2000), dtype='int')
    bpm2 = ccdmask(flat1, flat2, mask, mode='full')
    assert bpm2[1].sum() == 100


def test_ccdmask_mask():
    flat1 = numpy.ones((2000, 2000))
    flat1[0:100, 0] = 0
    flat2 = flat1
    mask = numpy.zeros((2000, 2000), dtype='int')
    mask[0:10, 0] = 0
    bpm2 = ccdmask(flat1, flat2, mask, mode='full')
    assert bpm2[1].sum() == 100



#
# Copyright 2014-2023 Universidad Complutense de Madrid
#
# This file is part of Numina
#
# SPDX-License-Identifier: GPL-3.0-or-later
# License-Filename: LICENSE.txt
#

import pytest

import numpy as np
from numpy.testing import assert_array_equal

from ..utils import coor_to_pix_1d, wcs_to_pix_np
from ..utils import slice_create, expand_slice, expand_region


def test_wctopix():

    xin = [0, 1, 2, 3, 4, 5]
    cpix = [coor_to_pix_1d(x) for x in xin]
    pix = [0, 1, 2, 3, 4, 5]

    assert_array_equal(cpix, pix)

    xin = [0.1, 1.6, 2.8, 3.5, 3.500001, 4.9999, 5.7]
    cpix = [coor_to_pix_1d(x) for x in xin]
    pix = [0, 2, 3, 4, 4, 5, 6]

    assert_array_equal(cpix, pix)


@pytest.mark.skipif(reason='Not clear if this is the interface')
def test_wctopixnp():

    xin = np.array([[0, 1], [2, 3], [4, 5]])
    cpix = wcs_to_pix_np(xin)
    pix = np.array([[0, 1], [2, 3], [4, 5]])

    assert_array_equal(cpix, pix)

    xin = [0.1, 1.6, 2.8, 3.5, 3.500001, 4.9999, 5.7]
    cpix = wcs_to_pix_np(xin)
    pix = [0, 2, 3, 4, 4, 5, 6]

    assert_array_equal(cpix, pix)


def test_slice_create():

    val = slice(97, 104, 1)
    cval = slice_create(100, 3)
    assert val == cval

    # check limits
    cval = slice_create(10, 12)
    resval = slice(0, 23, 1)
    assert resval == cval

    cval = slice_create(100, 12, stop=105)
    resval = slice(88, 105, 1)
    assert resval == cval


def test_expand_slice():
    s = slice(94, 104, 1)
    cval = expand_slice(s, 3, 4)
    assert slice(91, 108, 1) == cval

    s = slice(94, 104, 1)
    cval = expand_slice(s, 3, 4, stop=106)
    assert slice(91, 106, 1) == cval

    s = slice(3, 100, 1)
    cval = expand_slice(s, 6, 5)
    assert slice(0, 105, 1) == cval


def test_expand_region():
    t = (slice(94, 104, 1), slice(103, 250, 1))
    ct = expand_region(t, 4, 5)
    assert ct == (slice(90, 109, 1), slice(99, 255, 1))
